/*
   Be a text-based xload

   $dodge: tl.c,v 1.21 2005/03/05 22:03:32 dodge Exp $
   $Name: version_3 $
*/

#include "tl.h"

// globals
time_t sleep_time;
WINDOW *win;
double *loadavg_history;
unsigned int columns_last;
int cursify;
int tl_reinit_win;

// Display the right way to use this program. If we got here, it's not
// good. Just exit.
void tl_usage (int argc, char **argv)
{
	fprintf(stderr, "Usage: %s [-s delay] [-c]\n", argv[0]);
	fprintf(stderr, "	-c		no curses, more debugging\n");
	fprintf(stderr, "	-d <filename>	send debugging output to filename\n");
	fprintf(stderr, "	-h		this stuff\n");
	fprintf(stderr, "	-s delay	time between updates\n");
	exit(1);
}

// wrapper for curses move. I think of this with 0,0 being at the bottom
// right, not the top left. I also think in terms of x, y and not y, x.
int tl_move (int x, int y)
{
	int maxx, maxy;

	getmaxyx(win, maxy, maxx);

	return(wmove(win, maxy - y - 1, maxx - x - 1));
}

// draw the screen
void draw_screen ( )
{
	int maxx, maxy, x, y, a, scale;
	char hostname[MAXHOSTNAMELEN];

	getmaxyx(win, maxy, maxx);

	debugout(__FILE__, __LINE__, "draw_screen 00 -- %dx%d\n", maxx, maxy);

	wclear(win);

	scale = 1;

	for (x = 0; x < maxx; x++)
	{
		if (scale <= loadavg_history[x]) {
			scale = loadavg_history[x] + 1;
		}
	}

	a = maxy / scale;

	for (x = 0; x < maxx; x++)
	{
		for (y = 0; y < maxy; y++)
		{
			if (loadavg_history[x] * maxy / y / scale > 1.0) {
				tl_move(x, y);
				wprintw(win, "*");
			}

			if (a > 3 && scale > 0)
			{
				// fewer than 4 lines of data looks ugly
				// (and might be wrong).
				if (a > 3 && y % a == a - 1)
				{
					tl_move(x, y);
					wprintw(win, "-");
				}
			}
		}
	}

	wmove(win, 0, 0);

	if (gethostname(hostname, MAXHOSTNAMELEN) == -1)
	{
		strcpy(hostname, "localhost");
	}

	wprintw(win, "%s: %0.2f ", hostname, loadavg_history[0]);
	if (a <= 3 || scale == 0)
	{
		printw("-- scale: %d ", scale);
	}

	wmove(win, 0, COLS - 1);

	refresh();
}

// the window has been resized. Do something about it here.
void tl_resize (int sigraised)
{
	int maxx, maxy;

	getmaxyx(win, maxy, maxx);
	
	debugout(__FILE__, __LINE__, "tl_resize 00 -- %dx%d\n", 
			maxx, maxy);

	tl_reinit_win = 1;
}

// Gets called when we get a signal. Cleanup the curses spooge and exit
// gracefully-ish
void tl_cleanup (int sigraised)
{
	debugout(__FILE__, __LINE__, "tl_cleanup 00\n");
	
	if (cursify)
	{
		if (endwin() == ERR)
		{
			fprintf(stderr, "Can't endwin: %s\n", strerror(errno));
			exit(1);
		}
	}

	if (sigraised != -1)
		fprintf(stderr, "Caught signal %d, cleaning up\n", sigraised);

	if (sigraised == SIGHUP || sigraised == SIGINT ||
		sigraised == SIGQUIT || sigraised == SIGKILL ||
		sigraised == SIGTERM || sigraised == -1)
	{
		exit(0);
	} else {
		abort();
	}
}

void tl_curses_version ( ) {

#ifdef SUN_OS_CLASS
	printf("curses: Sun\n");
#endif /* SUN_OS_CLASS */
#ifdef NCURSES_VERSION
	printf("curses: %s\n", curses_version());
#endif /* NCURSES_VERSION */

	exit(0);
}

// miscellaneous initialization stuff. Go through the args, setup the
// screen.
void tl_init (int argc, char **argv)
{
	int ch;

	loadavg_history = NULL;

	columns_last = 0;

	sleep_time = 5;

	cursify = 1;

	while ((ch = getopt(argc, argv, "cCd:h?s:")) != -1)
	{
		switch (ch) {
		case 'c':
			errno = 0;
			cursify = 0;
			COLS = 1;
			break;

		case 'C':
			tl_curses_version();
			break;

		case 'd':
			debug_filename = (char *)optarg;
			break;

		case 'h':
				tl_usage(argc, argv);
				break;

		case '?':
				tl_usage(argc, argv);
				break;

		case 's':
			errno = 0;
			sleep_time = strtol(optarg, NULL, 0);
			break;

		default:
			tl_usage(argc, argv);
		}
	}

	// Trap signals to cleanup as much as possible
	for (ch = 0; ch < 256; ch++)
	{
		if (ch != SIGABRT && ch != SIGWINCH)
		{
			signal(ch, &tl_cleanup);
		}
	}

	// trap the window-change signal
	signal(SIGWINCH, &tl_resize);

	if (cursify)
	{
		// Setup the screen
		if ((win = initscr()) == NULL)
		{
			fprintf(stderr, "Can't initscr: %s\n", strerror(errno));
			exit(1);
		}
		nodelay(win, TRUE);
	}
}



// Update the array of list of load averages. Most recent is at [0].
void tl_add_load(double loadavg)
{
	unsigned int columns_now, i;
	double *new_loadavg_history;

	debugout(__FILE__, __LINE__, "tl_add_load 00 -- %dx%d\n", COLS, LINES);

	columns_now = cursify ? COLS : columns_last + 1;

	if (columns_now != columns_last)
	{
		// screen has changed size. Expand or contract the array,
		// then copy the old stuff in, then free the old stuff. 
		if ((new_loadavg_history = malloc(columns_now * sizeof(loadavg))) == NULL)
		{
			fprintf(stderr, "Can't malloc %d bytes: %s\n",
				columns_now * sizeof(loadavg),
				strerror(errno));
			exit(1);
		}

		bzero(new_loadavg_history, columns_now * sizeof(loadavg));

		new_loadavg_history[0] = loadavg;

		if (loadavg_history != NULL)
		{
			for (i = 1; i <= columns_now - 1 && i <= columns_last - 1; i++)
			{
				new_loadavg_history[i] = loadavg_history[i-1];
			}

// segfaults when screen trimmed down. wtf?
//			free(loadavg_history);
		}

		loadavg_history = new_loadavg_history;

		columns_last = columns_now;

		// I should probably call resizeterm from drawscreen(), but
		// I know to do it here.
		wresize(win, LINES, COLS);
	} else {
		// array would be the same size. Just move to the back of
		// the bus to let the new folks on.
		for (i = columns_last; i > 0; i--)
		{
			loadavg_history[i] = loadavg_history[i-1];
		}
		loadavg_history[0] = loadavg;
	}

	if (!cursify)
	{
		for (i = 0; i < columns_last; i++)
		{
			printf("%d: %f\n", i, loadavg_history[i]);
		}
		printf("\n");
	}
}


int good_ncurses[] = {
	20010224,	// OpenBSD
	20020209,	// Mac OS X
	0
};
int ncurses_version_hack ( )
{
	// There's a problem with older versions of ncurses that cause
	// segfaults when resizing xterms. The point here is to prevent
	// resizing on systems with such libraries.
	const char *cv, *p;
	unsigned int pl, i;

#ifndef NCURSES_VERSION
	return(1);			// non-ncurses seems to work.

#else //NCURSES_VERSION

	cv = curses_version();

	if ((p = rindex(cv, '.')) == NULL)
	{
		return(1);
	}

	pl = strtoul(p + 1, NULL, 0);

	for (i = 0; good_ncurses[i] != 0; i++)
		if (good_ncurses[i] == pl)
			return(1);

	return(0);
#endif //NCURSES_VERSION
}


int main (int argc, char **argv)
{
	double loadavg[3];
	int maxy, maxx, i, retval, loadavg_to_use;

	tl_init(argc, argv);
	debuginit();
	tl_reinit_win = 0;
	getmaxyx(win, maxy, maxx);
	while (1)
	{
		getloadavg(loadavg, 3);
		debugout(__FILE__, __LINE__, "main 10 -- %dx%d -- %d\n",
				maxx, maxy, tl_reinit_win);
		if (tl_reinit_win && ncurses_version_hack())
		{
			endwin();
			win = initscr();
			nodelay(win, TRUE);
			getmaxyx(win, maxx, maxy);
			debugout(__FILE__, __LINE__, "main 11: %dx%d\n",
					maxx, maxy);
			tl_reinit_win = 0;
			draw_screen();
		}

		if (sleep_time < 300)
		{
			loadavg_to_use = 0;
		} else if (sleep_time < 900)
		{
			loadavg_to_use = 1;
		} else 
		{
			loadavg_to_use = 2;
		}

		tl_add_load(loadavg[loadavg_to_use]);
		draw_screen();
		for (i = 0; i < sleep_time; i++)
		{
			sleep(1);

			if ((retval = getch()) != ERR)
			{
				if ((char)retval == 'q')
				{
					tl_cleanup(-1);
				}
			}
		}
	}
}
